use crate::entities::{prelude::*, *};
use crate::state;
use crate::{error::AppError, views};
use argon2::password_hash::rand_core::OsRng;
use argon2::password_hash::SaltString;
use argon2::{Argon2, PasswordHasher};
use axum::error_handling::HandleErrorLayer;
use axum::extract::State;
use axum::{async_trait, BoxError};
use axum::{http::StatusCode, response::Html, routing::get, Form, Router};
use axum_csrf::{CsrfConfig, CsrfLayer, CsrfToken};
use axum_login::tower_sessions::{MemoryStore, SessionManagerLayer};
use axum_login::{AuthManagerLayer, AuthUser, AuthnBackend, UserId};
use base64::prelude::*;
use maud::html;
use notify::Watcher;
use password_hash::{PasswordHash, PasswordVerifier};
use sea_orm::*;
use serde::Deserialize;
use std::{env, path::Path};
use tower::ServiceBuilder;
use tower_http::services::ServeDir;
use tower_livereload::LiveReloadLayer;
use tracing::{info, instrument};

#[instrument]
pub async fn new() -> Result<Router, anyhow::Error> {
    let app_router = Router::new()
        .route("/hello-world", get(views::greet_world))
        .route("/hello-world-text", get(views::greet_world_text));

    let assets_dir = ServeDir::new("./assets");
    let state = state::State::new().await?;

    let live_reload_layer = LiveReloadLayer::new();
    let reloader = live_reload_layer.reloader();

    let mut watcher = notify::recommended_watcher(move |_| reloader.reload())?;
    watcher.watch(Path::new("assets"), notify::RecursiveMode::Recursive)?;

    let cookie_key_bytes: Vec<u8> = BASE64_URL_SAFE_NO_PAD
        .decode(env::var("COOKIE_KEY").expect("COOKIE_KEY not set"))
        .expect("COOKIE_KEY not base64 URL-safe encoded");
    let cookie_key = cookie::Key::from(&cookie_key_bytes);
    let csrf_config = CsrfConfig::default().with_key(Some(cookie_key));

    let session_store = MemoryStore::default();
    let login_session_manager =
        SessionManagerLayer::new(session_store).with_name("login_sessions.sid");

    let auth_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|_: BoxError| async {
            StatusCode::BAD_REQUEST
        }))
        .layer(AuthManagerLayer::new(state.clone(), login_session_manager));

    let router = Router::new()
        .fallback(|| async { (StatusCode::NOT_FOUND, "404 page not found") })
        .nest("/app", app_router)
        .nest_service("/assets", assets_dir)
        .route("/", get(views::index))
        .route("/login", get(views::login).post(login))
        .route("/register", get(views::register).post(register))
        .route("/all_users", get(views::all_users))
        .with_state(state)
        .layer(CsrfLayer::new(csrf_config))
        .layer(auth_service)
        .layer(live_reload_layer);

    Ok(router)
}

fn csrf_verify(c: CsrfToken, t: &str) -> Result<(), AppError> {
    // TODO: https://docs.rs/axum_csrf/latest/axum_csrf/#prevent-post-replay-attacks-with-csrf
    c.verify(t)?;
    Ok(())
}

type AppRes = Result<(StatusCode, Html<String>), AppError>;

#[derive(Deserialize)]
struct Register {
    authenticity_token: String,
    username: String,
    password: String,
}

impl TryInto<user::ActiveModel> for Register {
    type Error = AppError;

    fn try_into(self) -> Result<user::ActiveModel, Self::Error> {
        Ok(user::ActiveModel {
            id: ActiveValue::Set(uuid::Uuid::now_v7().into()),
            name: ActiveValue::Set(None),
            username: ActiveValue::Set(self.username),
            password_digest: ActiveValue::Set(password_digest(self.password)?),
        })
    }
}

fn password_digest<S>(s: S) -> Result<String, password_hash::Error>
where
    S: AsRef<str>,
{
    Ok(Argon2::default()
        .hash_password(s.as_ref().as_bytes(), &SaltString::generate(&mut OsRng))?
        .serialize()
        .to_string())
}

fn password_verify<S>(password: S, current_digest: S) -> Result<(), password_hash::Error>
where
    S: AsRef<str>,
{
    let password_bytes = password.as_ref().as_bytes();
    let current = PasswordHash::new(current_digest.as_ref())?;
    Ok(Argon2::default().verify_password(password_bytes, &current)?)
}

impl user::ActiveModel {}

async fn register(c: CsrfToken, State(s): State<state::State>, Form(f): Form<Register>) -> AppRes {
    csrf_verify(c, &f.authenticity_token)?;

    // TODO: handle duplicate username

    let new: user::ActiveModel = f.try_into()?;
    let res = User::insert(new).exec(&s.db).await;

    info!("insert new user result: {:?}", res);

    res.map_err(anyhow::Error::from)?;

    Ok((
        StatusCode::CREATED,
        Html(
            html! {
                // h1 { (new_user.username) }
            }
            .into_string(),
        ),
    ))
}

#[derive(Deserialize, Clone)]
pub struct Login {
    authenticity_token: String,
    username: String,
    password: String,
}

type AuthSession = axum_login::AuthSession<state::State>;

impl AuthUser for user::Model {
    type Id = uuid::Uuid;

    fn id(&self) -> Self::Id {
        uuid::Uuid::try_from(self.id.clone()).expect("failed to convert user ID to UUID")
    }

    fn session_auth_hash(&self) -> &[u8] {
        self.password_digest.as_bytes()
    }
}

#[async_trait]
impl AuthnBackend for state::State {
    type User = user::Model;
    type Credentials = Login;
    type Error = AppError;

    async fn authenticate(&self, l: Self::Credentials) -> Result<Option<Self::User>, Self::Error> {
        match User::find()
            .filter(user::Column::Username.eq(l.username))
            // TODO: will this have index problems since I'm searching over the password digest?
            .one(&self.db)
            .await?
        {
            Some(user) => match password_verify(&l.password, &user.password_digest) {
                Ok(()) => Ok(Some(user)),
                Err(e) => Err(e.into()),
            },
            None => Ok(None),
        }
    }

    async fn get_user(&self, user_id: &UserId<Self>) -> Result<Option<Self::User>, Self::Error> {
        Ok(User::find_by_id(*user_id).one(&self.db).await?)
    }
}

async fn login(mut auth: AuthSession, c: CsrfToken, Form(f): Form<Login>) -> AppRes {
    csrf_verify(c, &f.authenticity_token)?;

    let user = match auth.authenticate(f.clone()).await {
        Ok(Some(user)) => user,
        Ok(None) => return Ok((StatusCode::UNAUTHORIZED, Html("user not found".to_string()))),
        Err(e) => {
            return Ok((
                StatusCode::INTERNAL_SERVER_ERROR,
                Html(format!("failed to authenticate user: {}", e)),
            ))
        }
    };

    if let Err(e) = auth.login(&user).await {
        return Ok((
            StatusCode::INTERNAL_SERVER_ERROR,
            Html(format!("failed to login user: {}", e)),
        ));
    }

    Ok((
        StatusCode::OK,
        Html(
            html! {
                h1 { (f.username) }
                h1 { (f.password) }
            }
            .into_string(),
        ),
    ))
}
